package com.sun.corba.se.impl.oa.poa;

import java.util.Set;
import org.omg.PortableInterceptor.ObjectReferenceTemplate;
import java.util.HashSet;
import org.omg.PortableServer.ForwardRequest;
import com.sun.corba.se.spi.protocol.ForwardException;
import com.sun.corba.se.spi.oa.OAInvocationInfo;
import com.sun.corba.se.spi.oa.OADestroyed;
import com.sun.corba.se.spi.copyobject.ObjectCopierFactory;
import org.omg.PortableServer.POAPackage.ServantNotActive;
import org.omg.PortableServer.POAPackage.ObjectNotActive;
import org.omg.PortableServer.POAPackage.ServantAlreadyActive;
import org.omg.PortableServer.POAPackage.ObjectAlreadyActive;
import org.omg.PortableServer.POAPackage.NoServant;
import org.omg.PortableServer.Servant;
import org.omg.PortableServer.POAPackage.WrongPolicy;
import org.omg.PortableServer.ServantManager;
import java.util.Collection;
import org.omg.PortableServer.RequestProcessingPolicy;
import org.omg.PortableServer.RequestProcessingPolicyValue;
import org.omg.PortableServer.ServantRetentionPolicy;
import org.omg.PortableServer.ServantRetentionPolicyValue;
import org.omg.PortableServer.ImplicitActivationPolicy;
import org.omg.PortableServer.ImplicitActivationPolicyValue;
import org.omg.PortableServer.IdAssignmentPolicy;
import org.omg.PortableServer.IdAssignmentPolicyValue;
import org.omg.PortableServer.IdUniquenessPolicy;
import org.omg.PortableServer.IdUniquenessPolicyValue;
import org.omg.PortableServer.LifespanPolicy;
import org.omg.PortableServer.LifespanPolicyValue;
import org.omg.PortableServer.ThreadPolicy;
import org.omg.PortableServer.ThreadPolicyValue;
import org.omg.CORBA.SystemException;
import org.omg.PortableServer.POAPackage.AdapterNonExistent;
import org.omg.PortableServer.POAPackage.InvalidPolicy;
import org.omg.PortableServer.POAPackage.AdapterAlreadyExists;
import org.omg.CORBA.Policy;
import org.omg.PortableServer.POAManager;
import java.util.Iterator;
import com.sun.corba.se.spi.ior.IOR;
import org.omg.PortableServer.POAPackage.WrongAdapter;
import com.sun.corba.se.spi.ior.TaggedProfile;
import com.sun.corba.se.spi.ior.IORFactories;
import com.sun.corba.se.impl.ior.POAObjectKeyTemplate;
import com.sun.corba.se.impl.orbutil.concurrent.ReentrantMutex;
import com.sun.corba.se.impl.ior.ObjectAdapterIdArray;
import java.util.HashMap;
import com.sun.corba.se.impl.orbutil.ORBUtility;
import com.sun.corba.se.impl.orbutil.concurrent.SyncUtil;
import com.sun.corba.se.spi.orb.ORB;
import com.sun.corba.se.impl.orbutil.concurrent.CondVar;
import com.sun.corba.se.impl.orbutil.concurrent.Sync;
import org.omg.PortableServer.AdapterActivator;
import java.util.Map;
import com.sun.corba.se.spi.ior.ObjectAdapterId;
import org.omg.PortableServer.POA;
import com.sun.corba.se.spi.oa.ObjectAdapterBase;

public class POAImpl extends ObjectAdapterBase implements POA
{
    private boolean debug;
    private static final int STATE_START = 0;
    private static final int STATE_INIT = 1;
    private static final int STATE_INIT_DONE = 2;
    private static final int STATE_RUN = 3;
    private static final int STATE_DESTROYING = 4;
    private static final int STATE_DESTROYED = 5;
    private int state;
    private POAPolicyMediator mediator;
    private int numLevels;
    private ObjectAdapterId poaId;
    private String name;
    private POAManagerImpl manager;
    private int uniquePOAId;
    private POAImpl parent;
    private Map children;
    private AdapterActivator activator;
    private int invocationCount;
    Sync poaMutex;
    private CondVar adapterActivatorCV;
    private CondVar invokeCV;
    private CondVar beingDestroyedCV;
    protected ThreadLocal isDestroying;
    
    private String stateToString() {
        switch (this.state) {
            case 0: {
                return "START";
            }
            case 1: {
                return "INIT";
            }
            case 2: {
                return "INIT_DONE";
            }
            case 3: {
                return "RUN";
            }
            case 4: {
                return "DESTROYING";
            }
            case 5: {
                return "DESTROYED";
            }
            default: {
                return "UNKNOWN(" + this.state + ")";
            }
        }
    }
    
    public String toString() {
        return "POA[" + this.poaId.toString() + ", uniquePOAId=" + this.uniquePOAId + ", state=" + this.stateToString() + ", invocationCount=" + this.invocationCount + "]";
    }
    
    boolean getDebug() {
        return this.debug;
    }
    
    static POAFactory getPOAFactory(ORB orb) {
        return (POAFactory)orb.getRequestDispatcherRegistry().getObjectAdapterFactory(32);
    }
    
    static POAImpl makeRootPOA(ORB orb) {
        POAManagerImpl poaManagerImpl = new POAManagerImpl(getPOAFactory(orb), orb.getPIHandler());
        POAImpl poaImpl = new POAImpl("RootPOA", null, orb, 0);
        poaImpl.initialize(poaManagerImpl, Policies.rootPOAPolicies);
        return poaImpl;
    }
    
    int getPOAId() {
        return this.uniquePOAId;
    }
    
    void lock() {
        SyncUtil.acquire(this.poaMutex);
        if (this.debug) {
            ORBUtility.dprint(this, "LOCKED poa " + this);
        }
    }
    
    void unlock() {
        if (this.debug) {
            ORBUtility.dprint(this, "UNLOCKED poa " + this);
        }
        this.poaMutex.release();
    }
    
    Policies getPolicies() {
        return this.mediator.getPolicies();
    }
    
    private POAImpl(String s, POAImpl poaImpl, ORB orb, int state) {
        super(orb);
        this.debug = orb.poaDebugFlag;
        if (this.debug) {
            ORBUtility.dprint(this, "Creating POA with name=" + s + " parent=" + poaImpl);
        }
        this.state = state;
        this.name = s;
        this.parent = poaImpl;
        this.children = new HashMap();
        this.activator = null;
        this.uniquePOAId = getPOAFactory(orb).newPOAId();
        if (poaImpl == null) {
            this.numLevels = 1;
        }
        else {
            this.numLevels = poaImpl.numLevels + 1;
            poaImpl.children.put(s, this);
        }
        String[] array = new String[this.numLevels];
        int n = this.numLevels - 1;
        while (this != null) {
            array[n--] = this.name;
            this = this.parent;
        }
        this.poaId = new ObjectAdapterIdArray(array);
        this.invocationCount = 0;
        this.poaMutex = new ReentrantMutex(orb.poaConcurrencyDebugFlag);
        this.adapterActivatorCV = new CondVar(this.poaMutex, orb.poaConcurrencyDebugFlag);
        this.invokeCV = new CondVar(this.poaMutex, orb.poaConcurrencyDebugFlag);
        this.beingDestroyedCV = new CondVar(this.poaMutex, orb.poaConcurrencyDebugFlag);
        this.isDestroying = new ThreadLocal() {
            @Override
            protected Object initialValue() {
                return Boolean.FALSE;
            }
        };
    }
    
    private void initialize(POAManagerImpl poaManagerImpl, Policies obj) {
        if (this.debug) {
            ORBUtility.dprint(this, "Initializing poa " + this + " with POAManager=" + poaManagerImpl + " policies=" + obj);
        }
        (this.manager = poaManagerImpl).addPOA(this);
        this.mediator = POAPolicyMediatorFactory.create(obj, this);
        POAObjectKeyTemplate obj2 = new POAObjectKeyTemplate(this.getORB(), this.mediator.getScid(), this.mediator.getServerId(), this.getORB().getORBData().getORBId(), this.poaId);
        if (this.debug) {
            ORBUtility.dprint(this, "Initializing poa: oktemp=" + obj2);
        }
        this.initializeTemplate(obj2, true, obj, null, null, obj2.getObjectAdapterId());
        if (this.state == 0) {
            this.state = 3;
        }
        else if (this.state == 1) {
            this.state = 2;
        }
        else {
            throw this.lifecycleWrapper().illegalPoaStateTrans();
        }
    }
    
    private boolean waitUntilRunning() {
        if (this.debug) {
            ORBUtility.dprint(this, "Calling waitUntilRunning on poa " + this);
        }
        while (this.state < 3) {
            try {
                this.adapterActivatorCV.await();
            }
            catch (InterruptedException ex) {}
        }
        if (this.debug) {
            ORBUtility.dprint(this, "Exiting waitUntilRunning on poa " + this);
        }
        return this.state == 3;
    }
    
    private boolean destroyIfNotInitDone() {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling destroyIfNotInitDone on poa " + this);
            }
            int n = this.state == 2;
            if (!(n == 0)) {
                this.state = 3;
            }
            else {
                new POAImpl.DestroyThread(false, this.debug).doIt(this, true);
            }
            return n;
        }
        finally {
            this.adapterActivatorCV.broadcast();
            if (this.debug) {
                ORBUtility.dprint(this, "Exiting destroyIfNotInitDone on poa " + this);
            }
            this.unlock();
        }
    }
    
    private byte[] internalReferenceToId(org.omg.CORBA.Object object) throws WrongAdapter {
        IOR ior = ORBUtility.getIOR(object);
        if (!IORFactories.getIORTemplateList(this.getCurrentFactory()).isEquivalent(ior.getIORTemplates())) {
            throw new WrongAdapter();
        }
        Iterator<Object> iterator = ior.iterator();
        if (!iterator.hasNext()) {
            throw this.iorWrapper().noProfilesInIor();
        }
        return ((TaggedProfile)(TaggedProfile)iterator.next()).getObjectId().getId();
    }
    
    void etherealizeAll() {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling etheralizeAll on poa " + this);
            }
            this.mediator.etherealizeAll();
        }
        finally {
            if (this.debug) {
                ORBUtility.dprint(this, "Exiting etheralizeAll on poa " + this);
            }
            this.unlock();
        }
    }
    
    public POA create_POA(String str, POAManager obj, Policy[] obj2) throws AdapterAlreadyExists, InvalidPolicy {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling create_POA(name=" + str + " theManager=" + obj + " policies=" + obj2 + ") on poa " + this);
            }
            if (this.state > 3) {
                throw this.omgLifecycleWrapper().createPoaDestroy();
            }
            POAImpl obj3 = (POAImpl)(POAImpl)this.children.get(str);
            if (obj3 == null) {
                obj3 = new POAImpl(str, this, this.getORB(), 0);
            }
            try {
                obj3.lock();
                if (this.debug) {
                    ORBUtility.dprint(this, "Calling create_POA: new poa is " + obj3);
                }
                if (obj3.state != 0 && obj3.state != 1) {
                    throw new AdapterAlreadyExists();
                }
                POAManagerImpl poaManagerImpl = (POAManagerImpl)obj;
                if (poaManagerImpl == null) {
                    poaManagerImpl = new POAManagerImpl(this.manager.getFactory(), this.manager.getPIHandler());
                }
                obj3.initialize(poaManagerImpl, new Policies(obj2, this.getORB().getCopierManager().getDefaultId()));
                obj3.unlock();
                return obj3;
            }
            finally {
                obj3.unlock();
            }
        }
        finally {
            this.unlock();
        }
    }
    
    public POA find_POA(String str, boolean b) throws AdapterNonExistent {
        AdapterActivator activator = null;
        this.lock();
        if (this.debug) {
            ORBUtility.dprint(this, "Calling find_POA(name=" + str + " activate=" + b + ") on poa " + this);
        }
        POAImpl poaImpl = (Object)this.children.get(str);
        if (poaImpl != null) {
            if (this.debug) {
                ORBUtility.dprint(this, "Calling find_POA: found poa " + poaImpl);
            }
            try {
                poaImpl.lock();
                this.unlock();
                if (!poaImpl.waitUntilRunning()) {
                    throw this.omgLifecycleWrapper().poaDestroyed();
                }
            }
            finally {
                poaImpl.unlock();
            }
        }
        else {
            try {
                if (this.debug) {
                    ORBUtility.dprint(this, "Calling find_POA: no poa found");
                }
                if (b && this.activator != null) {
                    poaImpl = new POAImpl(str, this, this.getORB(), 1);
                    if (this.debug) {
                        ORBUtility.dprint(this, "Calling find_POA: created poa " + poaImpl);
                    }
                    activator = this.activator;
                }
                else {
                    throw new AdapterNonExistent();
                }
            }
            finally {
                this.unlock();
            }
        }
        if (activator != null) {
            boolean unknown_adapter = false;
            Block_7: {
                if (this.debug) {
                    continue;
                }
                ORBUtility.dprint(this, "Calling find_POA: calling AdapterActivator");
                int destroyIfNotInitDone;
                try {
                    synchronized (activator) {
                        unknown_adapter = activator.unknown_adapter(this, str);
                    }
                    poaImpl.destroyIfNotInitDone();
                }
                catch (SystemException ex) {
                    throw this.omgLifecycleWrapper().adapterActivatorException(ex, str, this.poaId.toString());
                }
                catch (Throwable t) {
                    this.lifecycleWrapper().unexpectedException(t, this.toString());
                    if ((t instanceof ThreadDeath)) {
                        throw (ThreadDeath)t;
                    }
                    poaImpl.destroyIfNotInitDone();
                }
                finally {
                    destroyIfNotInitDone = poaImpl.destroyIfNotInitDone();
                }
                if (!(unknown_adapter == 0)) {
                    if (destroyIfNotInitDone == 0) {
                        throw this.omgLifecycleWrapper().adapterActivatorException(str, this.poaId.toString());
                    }
                }
                else {
                    if (this.debug) {
                        ORBUtility.dprint(this, "Calling find_POA: AdapterActivator returned false");
                    }
                    throw new AdapterNonExistent();
                }
            }
        }
        return poaImpl;
    }
    
    public void destroy(boolean b, boolean b2) {
        if (b2 && this.getORB().isDuringDispatch()) {
            throw this.lifecycleWrapper().destroyDeadlock();
        }
        new POAImpl.DestroyThread(b, this.debug).doIt(this, b2);
        return;
    }
    
    public ThreadPolicy create_thread_policy(ThreadPolicyValue threadPolicyValue) {
        return new ThreadPolicyImpl(threadPolicyValue);
    }
    
    public LifespanPolicy create_lifespan_policy(LifespanPolicyValue lifespanPolicyValue) {
        return new LifespanPolicyImpl(lifespanPolicyValue);
    }
    
    public IdUniquenessPolicy create_id_uniqueness_policy(IdUniquenessPolicyValue idUniquenessPolicyValue) {
        return new IdUniquenessPolicyImpl(idUniquenessPolicyValue);
    }
    
    public IdAssignmentPolicy create_id_assignment_policy(IdAssignmentPolicyValue idAssignmentPolicyValue) {
        return new IdAssignmentPolicyImpl(idAssignmentPolicyValue);
    }
    
    public ImplicitActivationPolicy create_implicit_activation_policy(ImplicitActivationPolicyValue implicitActivationPolicyValue) {
        return new ImplicitActivationPolicyImpl(implicitActivationPolicyValue);
    }
    
    public ServantRetentionPolicy create_servant_retention_policy(ServantRetentionPolicyValue servantRetentionPolicyValue) {
        return new ServantRetentionPolicyImpl(servantRetentionPolicyValue);
    }
    
    public RequestProcessingPolicy create_request_processing_policy(RequestProcessingPolicyValue requestProcessingPolicyValue) {
        return new RequestProcessingPolicyImpl(requestProcessingPolicyValue);
    }
    
    public String the_name() {
        try {
            this.lock();
            return this.name;
        }
        finally {
            this.unlock();
        }
    }
    
    public POA the_parent() {
        try {
            this.lock();
            return this.parent;
        }
        finally {
            this.unlock();
        }
    }
    
    public POA[] the_children() {
        try {
            this.lock();
            Collection<Object> values = this.children.values();
            POA[] array = new POA[values.size()];
            int n = 0;
            Iterator<Object> iterator = values.iterator();
            while (iterator.hasNext()) {
                array[n++] = (POA)(POA)iterator.next();
            }
            return array;
        }
        finally {
            this.unlock();
        }
    }
    
    public POAManager the_POAManager() {
        try {
            this.lock();
            return this.manager;
        }
        finally {
            this.unlock();
        }
    }
    
    public AdapterActivator the_activator() {
        try {
            this.lock();
            return this.activator;
        }
        finally {
            this.unlock();
        }
    }
    
    public void the_activator(AdapterActivator adapterActivator) {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling the_activator on poa " + this + " activator=" + adapterActivator);
            }
            this.activator = adapterActivator;
        }
        finally {
            this.unlock();
        }
    }
    
    public ServantManager get_servant_manager() throws WrongPolicy {
        try {
            this.lock();
            return this.mediator.getServantManager();
        }
        finally {
            this.unlock();
        }
    }
    
    public void set_servant_manager(ServantManager servantManager) throws WrongPolicy {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling set_servant_manager on poa " + this + " servantManager=" + servantManager);
            }
            this.mediator.setServantManager(servantManager);
        }
        finally {
            this.unlock();
        }
    }
    
    public Servant get_servant() throws NoServant, WrongPolicy {
        try {
            this.lock();
            return this.mediator.getDefaultServant();
        }
        finally {
            this.unlock();
        }
    }
    
    public void set_servant(Servant servant) throws WrongPolicy {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling set_servant on poa " + this + " defaultServant=" + servant);
            }
            this.mediator.setDefaultServant(servant);
        }
        finally {
            this.unlock();
        }
    }
    
    public byte[] activate_object(Servant obj) throws ServantAlreadyActive, WrongPolicy {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling activate_object on poa " + this + " (servant=" + obj + ")");
            }
            byte[] systemId = this.mediator.newSystemId();
            try {
                this.mediator.activateObject(systemId, obj);
            }
            catch (ObjectAlreadyActive objectAlreadyActive) {}
            return systemId;
        }
        finally {
            if (this.debug) {
                ORBUtility.dprint(this, "Exiting activate_object on poa " + this);
            }
            this.unlock();
        }
    }
    
    public void activate_object_with_id(byte[] obj, Servant obj2) throws ObjectAlreadyActive, ServantAlreadyActive, WrongPolicy {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling activate_object_with_id on poa " + this + " (servant=" + obj2 + " id=" + obj + ")");
            }
            this.mediator.activateObject((byte[])(byte[])obj.clone(), obj2);
        }
        finally {
            if (this.debug) {
                ORBUtility.dprint(this, "Exiting activate_object_with_id on poa " + this);
            }
            this.unlock();
        }
    }
    
    public void deactivate_object(byte[] obj) throws ObjectNotActive, WrongPolicy {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling deactivate_object on poa " + this + " (id=" + obj + ")");
            }
            this.mediator.deactivateObject(obj);
        }
        finally {
            if (this.debug) {
                ORBUtility.dprint(this, "Exiting deactivate_object on poa " + this);
            }
            this.unlock();
        }
    }
    
    public org.omg.CORBA.Object create_reference(String str) throws WrongPolicy {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling create_reference(repId=" + str + ") on poa " + this);
            }
            return this.makeObject(str, this.mediator.newSystemId());
        }
        finally {
            this.unlock();
        }
    }
    
    public org.omg.CORBA.Object create_reference_with_id(byte[] obj, String str) {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling create_reference_with_id(oid=" + obj + " repId=" + str + ") on poa " + this);
            }
            return this.makeObject(str, (byte[])(byte[])obj.clone());
        }
        finally {
            this.unlock();
        }
    }
    
    public byte[] servant_to_id(Servant obj) throws ServantNotActive, WrongPolicy {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling servant_to_id(servant=" + obj + ") on poa " + this);
            }
            return this.mediator.servantToId(obj);
        }
        finally {
            this.unlock();
        }
    }
    
    public org.omg.CORBA.Object servant_to_reference(Servant obj) throws ServantNotActive, WrongPolicy {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling servant_to_reference(servant=" + obj + ") on poa " + this);
            }
            byte[] servantToId = this.mediator.servantToId(obj);
            return this.create_reference_with_id(servantToId, obj._all_interfaces(this, servantToId)[0]);
        }
        finally {
            this.unlock();
        }
    }
    
    public Servant reference_to_servant(org.omg.CORBA.Object obj) throws ObjectNotActive, WrongPolicy, WrongAdapter {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling reference_to_servant(reference=" + obj + ") on poa " + this);
            }
            if (this.state >= 4) {
                throw this.lifecycleWrapper().adapterDestroyed();
            }
            return this.mediator.idToServant(this.internalReferenceToId(obj));
        }
        finally {
            this.unlock();
        }
    }
    
    public byte[] reference_to_id(org.omg.CORBA.Object obj) throws WrongAdapter, WrongPolicy {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling reference_to_id(reference=" + obj + ") on poa " + this);
            }
            if (this.state >= 4) {
                throw this.lifecycleWrapper().adapterDestroyed();
            }
            return this.internalReferenceToId(obj);
        }
        finally {
            this.unlock();
        }
    }
    
    public Servant id_to_servant(byte[] obj) throws ObjectNotActive, WrongPolicy {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling id_to_servant(id=" + obj + ") on poa " + this);
            }
            if (this.state >= 4) {
                throw this.lifecycleWrapper().adapterDestroyed();
            }
            return this.mediator.idToServant(obj);
        }
        finally {
            this.unlock();
        }
    }
    
    public org.omg.CORBA.Object id_to_reference(byte[] obj) throws ObjectNotActive, WrongPolicy {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling id_to_reference(id=" + obj + ") on poa " + this);
            }
            if (this.state >= 4) {
                throw this.lifecycleWrapper().adapterDestroyed();
            }
            return this.makeObject(this.mediator.idToServant(obj)._all_interfaces(this, obj)[0], obj);
        }
        finally {
            this.unlock();
        }
    }
    
    public byte[] id() {
        try {
            this.lock();
            return this.getAdapterId();
        }
        finally {
            this.unlock();
        }
    }
    
    public Policy getEffectivePolicy(int n) {
        return this.mediator.getPolicies().get_effective_policy(n);
    }
    
    public int getManagerId() {
        return this.manager.getManagerId();
    }
    
    public short getState() {
        return this.manager.getORTState();
    }
    
    public String[] getInterfaces(Object o, byte[] array) {
        return ((Servant)o)._all_interfaces(this, array);
    }
    
    protected ObjectCopierFactory getObjectCopierFactory() {
        return this.getORB().getCopierManager().getObjectCopierFactory(this.mediator.getPolicies().getCopierId());
    }
    
    public void enter() throws OADestroyed {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling enter on poa " + this);
            }
            while (this.state == 4 && this.isDestroying.get() == Boolean.FALSE) {
                try {
                    this.beingDestroyedCV.await();
                }
                catch (InterruptedException ex) {}
            }
            if (!this.waitUntilRunning()) {
                throw new OADestroyed();
            }
            ++this.invocationCount;
        }
        finally {
            if (this.debug) {
                ORBUtility.dprint(this, "Exiting enter on poa " + this);
            }
            this.unlock();
        }
        this.manager.enter();
    }
    
    public void exit() {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling exit on poa " + this);
            }
            --this.invocationCount;
            if (this.invocationCount == 0 && this.state == 4) {
                this.invokeCV.broadcast();
            }
        }
        finally {
            if (this.debug) {
                ORBUtility.dprint(this, "Exiting exit on poa " + this);
            }
            this.unlock();
        }
        this.manager.exit();
    }
    
    public void getInvocationServant(OAInvocationInfo oaInvocationInfo) {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling getInvocationServant on poa " + this);
            }
            Object invocationServant;
            try {
                invocationServant = this.mediator.getInvocationServant(oaInvocationInfo.id(), oaInvocationInfo.getOperation());
            }
            catch (ForwardRequest forwardRequest) {
                throw new ForwardException(this.getORB(), forwardRequest.forward_reference);
            }
            oaInvocationInfo.setServant(invocationServant);
        }
        finally {
            if (this.debug) {
                ORBUtility.dprint(this, "Exiting getInvocationServant on poa " + this);
            }
            this.unlock();
        }
    }
    
    public org.omg.CORBA.Object getLocalServant(byte[] array) {
        return null;
    }
    
    public void returnServant() {
        try {
            this.lock();
            if (this.debug) {
                ORBUtility.dprint(this, "Calling returnServant on poa " + this);
            }
            this.mediator.returnServant();
        }
        catch (Throwable obj) {
            if (this.debug) {
                ORBUtility.dprint(this, "Exception " + obj + " in returnServant on poa " + this);
            }
            if ((obj instanceof Error)) {
                throw (Error)obj;
            }
            if ((obj instanceof RuntimeException)) {
                throw (RuntimeException)obj;
            }
            if (this.debug) {
                ORBUtility.dprint(this, "Exiting returnServant on poa " + this);
            }
            this.unlock();
        }
        finally {
            if (this.debug) {
                ORBUtility.dprint(this, "Exiting returnServant on poa " + this);
            }
            this.unlock();
        }
    }
    
    static class DestroyThread extends Thread
    {
        private boolean wait;
        private boolean etherealize;
        private boolean debug;
        private POAImpl thePoa;
        
        public DestroyThread(boolean etherealize, boolean debug) {
            this.etherealize = etherealize;
            this.debug = debug;
        }
        
        public void doIt(POAImpl poaImpl, boolean b) {
            if (this.debug) {
                ORBUtility.dprint(this, "Calling DestroyThread.doIt(thePOA=" + poaImpl + " wait=" + b + " etherealize=" + this.etherealize);
            }
            this.thePoa = poaImpl;
            this.wait = b;
            if (b) {
                this.run();
            }
            else {
                try {
                    this.setDaemon(true);
                }
                catch (Exception ex) {}
                this.start();
            }
        }
        
        @Override
        public void run() {
            HashSet<Object> set = new HashSet();
            this.performDestroy(this.thePoa, set);
            Iterator<Object> iterator = set.iterator();
            ObjectReferenceTemplate[] array = new ObjectReferenceTemplate[set.size()];
            int n = 0;
            while (iterator.hasNext()) {
                array[n++] = (ObjectReferenceTemplate)iterator.next();
            }
            this.thePoa.getORB().getPIHandler().adapterStateChanged(array, (short)4);
        }
        
        private boolean prepareForDestruction(POAImpl obj, Set set) {
            POAImpl[] array = null;
            try {
                obj.lock();
                if (this.debug) {
                    ORBUtility.dprint(this, "Calling performDestroy on poa " + obj);
                }
                if (obj.state <= 3) {
                    obj.state = 4;
                    obj.isDestroying.set(Boolean.TRUE);
                    array = (POAImpl[])(POAImpl[])obj.children.values().toArray(new POAImpl[0]);
                }
                else {
                    if (this.wait) {
                        while (obj.state != 5) {
                            try {
                                obj.beingDestroyedCV.await();
                            }
                            catch (InterruptedException ex) {}
                        }
                    }
                    return false;
                }
            }
            finally {
                obj.unlock();
            }
            for (int i = 0; i < array.length; ++i) {
                this.performDestroy(array[i], set);
            }
            return true;
        }
        
        public void performDestroy(POAImpl poaImpl, Set set) {
            if (!this.prepareForDestruction(poaImpl, set)) {
                return;
            }
            POAImpl access$300 = poaImpl.parent;
            int n = access$300 == null;
            try {
                if (n == 0) {
                    access$300.lock();
                }
                try {
                    poaImpl.lock();
                    this.completeDestruction(poaImpl, access$300, set);
                }
                finally {
                    poaImpl.unlock();
                    if (!(n == 0)) {
                        poaImpl.manager.getFactory().registerRootPOA();
                    }
                }
            }
            finally {
                if (n == 0) {
                    access$300.unlock();
                    poaImpl.parent = null;
                }
            }
            return;
        }
        
        private void completeDestruction(POAImpl obj, POAImpl poaImpl, Set set) {
            if (this.debug) {
                ORBUtility.dprint(this, "Calling completeDestruction on poa " + obj);
            }
            try {
                while (obj.invocationCount != 0) {
                    try {
                        obj.invokeCV.await();
                    }
                    catch (InterruptedException ex) {}
                }
                if (obj.mediator != null) {
                    if (this.etherealize) {
                        obj.mediator.etherealizeAll();
                    }
                    obj.mediator.clearAOM();
                }
                if (obj.manager != null) {
                    obj.manager.removePOA(obj);
                }
                if (poaImpl != null) {
                    poaImpl.children.remove(obj.name);
                }
                set.add(obj.getAdapterTemplate());
            }
            catch (Throwable t) {
                if ((t instanceof ThreadDeath)) {
                    throw (ThreadDeath)t;
                }
                obj.lifecycleWrapper().unexpectedException(t, obj.toString());
                obj.state = 5;
                obj.beingDestroyedCV.broadcast();
                obj.isDestroying.set(Boolean.FALSE);
                if (this.debug) {
                    ORBUtility.dprint(this, "Exiting completeDestruction on poa " + obj);
                }
            }
            finally {
                obj.state = 5;
                obj.beingDestroyedCV.broadcast();
                obj.isDestroying.set(Boolean.FALSE);
                if (this.debug) {
                    ORBUtility.dprint(this, "Exiting completeDestruction on poa " + obj);
                }
            }
        }
    }
}
